import numpy as np 
from scipy.special import gamma, kv
'''
This is a library to compute the correct confidence widths
for the different kernels. In particular, we use linear,
squared exponential, and RBF
'''

def matern_kernel(r, l = 1, v = 1):
    r = np.abs(r)
    r[r == 0] = 1e-8
    part1 = 2 ** (1 - v) / gamma(v)
    part2 = (np.sqrt(2 * v) * r / l) ** v
    part3 = kv(v, np.sqrt(2 * v) * r / l)
    return part1 * part2 * part3

def linear(B, delta, d, L, t, sig):
    '''
    confidence widths from Thm 20.5 in Tor and Csaba book
    B - scalar, bound on ||theta_ast||
    delta - scalar, conf probability
    d - dimension
    L - bound on ||x_i||, arm norm
    t - time step
    sig - subGaussian parameter
    '''
    return B + sig * np.sqrt(2 * np.log(1/delta) + d * np.log(1 + (t+1) * L**2 / d))

def sq_exponential(B, sig, delta, t, d):
    '''
    Confidence widths for squared exponential kernel. Using theoretical
    upper bound on information gain. 
    - B is bound on RKHS norm of theta: B = sqrt(kernel(theta, theta))
    - sig is the subgaussian paramter
    - delta is the confidence with
    - t is time step
    - d is dimension
    '''
    gamma_t = np.log((t+1)**(d + 1) * np.log(t+1))**(d + 1)
    return B + sig * np.sqrt(2 * (gamma_t + 1 + np.log(1/delta)))

def mattern(B, sig, delta, t, d, nu):
    '''
    Confidence widths for squared exponential kernel. Using theoretical
    upper bound on information gain. 
    - B is bound on RKHS norm of theta: B = sqrt(kernel(theta, theta))
    - sig is the subgaussian paramter
    - delta is the confidence with
    - t is time step
    - d is dimension
    - nu is roughness parameter of
    '''
    eta = nu / (2*nu + d * (d + 1))
    gamma_t = t**(1 - 2*eta) * np.log(t)
    return B + sig * np.sqrt(2 * (gamma_t + 1 + np.log(1/delta)))



